# TPK4186 - Advanced Tools for Performance Engineering
# IrrigationSystem.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. IrrigationSystemState
# 3. IrrigationSystem
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. IrrigationSystemStates
# -------------------------

class IrrigationSystemState(Alea.DiscreteEventSystems.State):
  CLOSED = 0
  OPEN = 1

  EMPTY = 0
  FULL = 1

  def __init__(self, irrigationSystem):
    Alea.DiscreteEventSystems.State.__init__(self, irrigationSystem)
    self.timeToFillTank1 = 12
    self.timeToEmptyTank1 = 20
    self.timeToFillTank2 = 10
    self.timeToEmptyTank2 = 15
    self.tank1 = IrrigationSystemState.EMPTY
    self.tank2 = IrrigationSystemState.EMPTY
    self.upstreamValveTank1 = IrrigationSystemState.CLOSED
    self.downstreamValveTank1 = IrrigationSystemState.CLOSED
    self.upstreamValveTank2 = IrrigationSystemState.CLOSED
    self.downstreamValveTank2 = IrrigationSystemState.CLOSED

  @staticmethod
  def New(irrigationSystem):
    return IrrigationSystemState(irrigationSystem)

  def Reset(self):
    self.tank1 = IrrigationSystemState.EMPTY
    self.tank2 = IrrigationSystemState.EMPTY
    self.upstreamValveTank1 = IrrigationSystemState.CLOSED
    self.downstreamValveTank1 = IrrigationSystemState.CLOSED
    self.upstreamValveTank2 = IrrigationSystemState.CLOSED
    self.downstreamValveTank2 = IrrigationSystemState.CLOSED

  def IsClientArrivalEnabled(self):
    return self.numberOfClientsWaiting<self.maximumSize

  def FireClientArrival(self):
    self.numberOfClientsWaiting += 1

  def IsStartClientServiceEnabled(self):
    return self.numberOfClientsWaiting>0 and self.numberOfClientsAtCounter==0

  def FireStartClientService(self):
    self.numberOfClientsWaiting -= 1
    self.numberOfClientsAtCounter += 1

  def IsCompleteClientServiceEnabled(self):
    return self.numberOfClientsAtCounter>0

  def FireCompleteClientService(self):
    self.numberOfClientsAtCounter -= 1

  def GetNumberOfClientsWaiting(self):
    return self.numberOfClientsWaiting

  def IsCounterIdle(self):
    if self.numberOfClientsAtCounter==0:
      return 1
    return 0

# 3. Bank
# -------

bank = Alea.DiscreteEventSystems.DiscreteEventSystem(BankState.New)
T1 = bank.NewTransition(BankState.IsClientArrivalEnabled, BankState.FireClientArrival, Alea.DiscreteEventSystems.ExponentialDelay(5))
T2 = bank.NewTransition(BankState.IsStartClientServiceEnabled, BankState.FireStartClientService, Alea.DiscreteEventSystems.DiracDelay(0))
T3 = bank.NewTransition(BankState.IsCompleteClientServiceEnabled, BankState.FireCompleteClientService, Alea.DiscreteEventSystems.TriangularDelay(3, 8, 5))

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(bank)
I1 = simulator.NewTransitionFiringCount("I1", T1, False)
I2 = simulator.NewTransitionFiringCount("I2", T3, False)
I3 = simulator.NewRewardMaximumValue("I3", BankState.GetNumberOfClientsWaiting, False)
I4 = simulator.NewRewardMeanValue("I4", BankState.GetNumberOfClientsWaiting, False)
I5 = simulator.NewRewardMeanValue("I5", BankState.IsCounterIdle, False)

simulator.MonteCarloSimulation(60, 10000)

print("I1\t{0:.2f}".format(I1.GetMean()))
print("I2\t{0:.2f}".format(I2.GetMean()))
print("I3\t{0:.2f}".format(I3.GetMaximumValue()))
print("I4\t{0:.2f}".format(I4.GetMean()))
print("I5\t{0:.2f}".format(I5.GetMean()))

