# TPK4186 - Advanced Tools for Performance Engineering
# PERProjectState.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Project States
# 3. Projects
# 4. Main

# 1. Required modules
# -------------------

import sys

import numpy as np
import matplotlib.pyplot as plt

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. Project States
# -----------------

class ProjectState(Alea.DiscreteEventSystems.State):
  TODO = 0
  STARTED = 1
  COMPLETED = 2

  def __init__(self, project):
    Alea.DiscreteEventSystems.State.__init__(self, project)
    self.activityS = ProjectState.COMPLETED
    self.activityA = ProjectState.TODO
    self.activityB = ProjectState.TODO
    self.activityC = ProjectState.TODO
    self.activityD = ProjectState.TODO
    self.activityE = ProjectState.TODO
    self.activityF = ProjectState.TODO
    self.activityG = ProjectState.TODO
    self.activityH = ProjectState.TODO
    self.activityI = ProjectState.TODO
    self.activityJ = ProjectState.TODO
    self.activityK = ProjectState.TODO
    self.activityL = ProjectState.TODO
    self.activityM = ProjectState.TODO
    self.activityN = ProjectState.TODO
    self.activityT = ProjectState.TODO

  @staticmethod
  def New(project):
    return ProjectState(project)

  def Reset(self):
    self.activityS = ProjectState.COMPLETED
    self.activityA = ProjectState.TODO
    self.activityB = ProjectState.TODO
    self.activityC = ProjectState.TODO
    self.activityD = ProjectState.TODO
    self.activityE = ProjectState.TODO
    self.activityF = ProjectState.TODO
    self.activityG = ProjectState.TODO
    self.activityH = ProjectState.TODO
    self.activityI = ProjectState.TODO
    self.activityJ = ProjectState.TODO
    self.activityK = ProjectState.TODO
    self.activityL = ProjectState.TODO
    self.activityM = ProjectState.TODO
    self.activityN = ProjectState.TODO
    self.activityT = ProjectState.TODO

  def CanActivityAStart(self):
    return self.activityA==ProjectState.TODO and self.activityS==ProjectState.COMPLETED
  def StartActivityA(self):
    self.activityA = ProjectState.STARTED
  def CanActivityABeCompleted(self):
    return self.activityA==ProjectState.STARTED
  def CompleteActivityA(self):
    self.activityA = ProjectState.COMPLETED

  def CanActivityBStart(self):
    return self.activityB==ProjectState.TODO and self.activityS==ProjectState.COMPLETED
  def StartActivityB(self):
    self.activityB = ProjectState.STARTED
  def CanActivityBBeCompleted(self):
    return self.activityB==ProjectState.STARTED
  def CompleteActivityB(self):
    self.activityB = ProjectState.COMPLETED

  def CanActivityCStart(self):
    return self.activityC==ProjectState.TODO and self.activityS==ProjectState.COMPLETED
  def StartActivityC(self):
    self.activityC = ProjectState.STARTED
  def CanActivityCBeCompleted(self):
    return self.activityC==ProjectState.STARTED
  def CompleteActivityC(self):
    self.activityC = ProjectState.COMPLETED

  def CanActivityDStart(self):
    return self.activityD==ProjectState.TODO and self.activityS==ProjectState.COMPLETED
  def StartActivityD(self):
    self.activityD = ProjectState.STARTED
  def CanActivityDBeCompleted(self):
    return self.activityD==ProjectState.STARTED
  def CompleteActivityD(self):
    self.activityD = ProjectState.COMPLETED

  def CanActivityEStart(self):
    return self.activityE==ProjectState.TODO and self.activityA==ProjectState.COMPLETED
  def StartActivityE(self):
    self.activityE = ProjectState.STARTED
  def CanActivityEBeCompleted(self):
    return self.activityE==ProjectState.STARTED
  def CompleteActivityE(self):
    self.activityE = ProjectState.COMPLETED

  def CanActivityFStart(self):
    return self.activityF==ProjectState.TODO and self.activityB==ProjectState.COMPLETED
  def StartActivityF(self):
    self.activityF = ProjectState.STARTED
  def CanActivityFBeCompleted(self):
    return self.activityF==ProjectState.STARTED
  def CompleteActivityF(self):
    self.activityF = ProjectState.COMPLETED

  def CanActivityGStart(self):
    return self.activityG==ProjectState.TODO and self.activityE==ProjectState.COMPLETED and self.activityF==ProjectState.COMPLETED
  def StartActivityG(self):
    self.activityG = ProjectState.STARTED
  def CanActivityGBeCompleted(self):
    return self.activityG==ProjectState.STARTED
  def CompleteActivityG(self):
    self.activityG = ProjectState.COMPLETED

  def CanActivityHStart(self):
    return self.activityH==ProjectState.TODO and self.activityG==ProjectState.COMPLETED
  def StartActivityH(self):
    self.activityH = ProjectState.STARTED
  def CanActivityHBeCompleted(self):
    return self.activityH==ProjectState.STARTED
  def CompleteActivityH(self):
    self.activityH = ProjectState.COMPLETED

  def CanActivityIStart(self):
    return self.activityI==ProjectState.TODO and self.activityC==ProjectState.COMPLETED and self.activityD==ProjectState.COMPLETED
  def StartActivityI(self):
    self.activityI = ProjectState.STARTED
  def CanActivityIBeCompleted(self):
    return self.activityI==ProjectState.STARTED
  def CompleteActivityI(self):
    self.activityI = ProjectState.COMPLETED

  def CanActivityJStart(self):
    return self.activityJ==ProjectState.TODO and self.activityI==ProjectState.COMPLETED
  def StartActivityJ(self):
    self.activityJ = ProjectState.STARTED
  def CanActivityJBeCompleted(self):
    return self.activityJ==ProjectState.STARTED
  def CompleteActivityJ(self):
    self.activityJ = ProjectState.COMPLETED

  def CanActivityKStart(self):
    return self.activityK==ProjectState.TODO and self.activityJ==ProjectState.COMPLETED
  def StartActivityK(self):
    self.activityK = ProjectState.STARTED
  def CanActivityKBeCompleted(self):
    return self.activityK==ProjectState.STARTED
  def CompleteActivityK(self):
    self.activityK = ProjectState.COMPLETED

  def CanActivityLStart(self):
    return self.activityL==ProjectState.TODO and self.activityI==ProjectState.COMPLETED
  def StartActivityL(self):
    self.activityL = ProjectState.STARTED
  def CanActivityLBeCompleted(self):
    return self.activityL==ProjectState.STARTED
  def CompleteActivityL(self):
    self.activityL = ProjectState.COMPLETED

  def CanActivityMStart(self):
    return self.activityM==ProjectState.TODO and self.activityL==ProjectState.COMPLETED
  def StartActivityM(self):
    self.activityM = ProjectState.STARTED
  def CanActivityMBeCompleted(self):
    return self.activityM==ProjectState.STARTED
  def CompleteActivityM(self):
    self.activityM = ProjectState.COMPLETED

  def CanActivityNStart(self):
    return self.activityN==ProjectState.TODO and self.activityM==ProjectState.COMPLETED
  def StartActivityN(self):
    self.activityN = ProjectState.STARTED
  def CanActivityNBeCompleted(self):
    return self.activityN==ProjectState.STARTED
  def CompleteActivityN(self):
    self.activityN = ProjectState.COMPLETED

  def CanProjectBeCompleted(self):
    return self.activityT==ProjectState.TODO and self.activityH==ProjectState.COMPLETED and self.activityK==ProjectState.COMPLETED and self.activityN==ProjectState.COMPLETED
  def CompleteProject(self):
    self.activityT = ProjectState.COMPLETED

# 3. Project
# ----------

project = Alea.DiscreteEventSystems.DiscreteEventSystem(ProjectState.New)
SA = project.NewTransition("StartActivityA", ProjectState.CanActivityAStart, ProjectState.StartActivityA, Alea.DiscreteEventSystems.DiracDelay(0))
CA = project.NewTransition("CompleteActivityA", ProjectState.CanActivityABeCompleted, ProjectState.CompleteActivityA, Alea.DiscreteEventSystems.TriangularDelay(59, 93, 70))

SB = project.NewTransition("StartActivityB", ProjectState.CanActivityBStart, ProjectState.StartActivityB, Alea.DiscreteEventSystems.DiracDelay(0))
CB = project.NewTransition("CompleteActivityB", ProjectState.CanActivityBBeCompleted, ProjectState.CompleteActivityB, Alea.DiscreteEventSystems.TriangularDelay(22, 56, 33))

SC = project.NewTransition("StartActivityC", ProjectState.CanActivityCStart, ProjectState.StartActivityC, Alea.DiscreteEventSystems.DiracDelay(0))
CC = project.NewTransition("CompleteActivityC", ProjectState.CanActivityCBeCompleted, ProjectState.CompleteActivityC, Alea.DiscreteEventSystems.TriangularDelay(22, 50, 30))

SD = project.NewTransition("StartActivityD", ProjectState.CanActivityDStart, ProjectState.StartActivityD, Alea.DiscreteEventSystems.DiracDelay(0))
CD = project.NewTransition("CompleteActivityD", ProjectState.CanActivityDBeCompleted, ProjectState.CompleteActivityD, Alea.DiscreteEventSystems.TriangularDelay(31, 61, 40))

SE = project.NewTransition("StartActivityE", ProjectState.CanActivityEStart, ProjectState.StartActivityE, Alea.DiscreteEventSystems.DiracDelay(0))
CE = project.NewTransition("CompleteActivityE", ProjectState.CanActivityEBeCompleted, ProjectState.CompleteActivityE, Alea.DiscreteEventSystems.TriangularDelay(30, 90, 51))

SF = project.NewTransition("StartActivityF", ProjectState.CanActivityFStart, ProjectState.StartActivityF, Alea.DiscreteEventSystems.DiracDelay(0))
CF = project.NewTransition(ProjectState.CanActivityFBeCompleted, ProjectState.CompleteActivityF, Alea.DiscreteEventSystems.TriangularDelay(22, 56, 33))

SG = project.NewTransition("StartActivityG", ProjectState.CanActivityGStart, ProjectState.StartActivityG, Alea.DiscreteEventSystems.DiracDelay(0))
CG = project.NewTransition("CompleteActivityG", ProjectState.CanActivityGBeCompleted, ProjectState.CompleteActivityG, Alea.DiscreteEventSystems.TriangularDelay(25, 51, 38))

SH = project.NewTransition("StartActivityH", ProjectState.CanActivityHStart, ProjectState.StartActivityH, Alea.DiscreteEventSystems.DiracDelay(0))
CH = project.NewTransition("CompleteActivityH", ProjectState.CanActivityHBeCompleted, ProjectState.CompleteActivityH, Alea.DiscreteEventSystems.TriangularDelay(36, 102, 54))

SI = project.NewTransition("StartActivityI", ProjectState.CanActivityIStart, ProjectState.StartActivityI, Alea.DiscreteEventSystems.DiracDelay(0))
CI = project.NewTransition("CompleteActivityI", ProjectState.CanActivityIBeCompleted, ProjectState.CompleteActivityI, Alea.DiscreteEventSystems.TriangularDelay(25, 55, 40))

SJ = project.NewTransition("StartActivityJ", ProjectState.CanActivityJStart, ProjectState.StartActivityJ, Alea.DiscreteEventSystems.DiracDelay(0))
CJ = project.NewTransition("CompleteActivityJ", ProjectState.CanActivityJBeCompleted, ProjectState.CompleteActivityJ, Alea.DiscreteEventSystems.TriangularDelay(59, 93, 70))

SK = project.NewTransition("StartActivityK", ProjectState.CanActivityKStart, ProjectState.StartActivityK, Alea.DiscreteEventSystems.DiracDelay(0))
CK = project.NewTransition("CompleteActivityK", ProjectState.CanActivityKBeCompleted, ProjectState.CompleteActivityK, Alea.DiscreteEventSystems.TriangularDelay(25, 45, 35))

SL = project.NewTransition("StartActivityL", ProjectState.CanActivityLStart, ProjectState.StartActivityL, Alea.DiscreteEventSystems.DiracDelay(0))
CL = project.NewTransition("CompleteActivityL", ProjectState.CanActivityLBeCompleted, ProjectState.CompleteActivityL, Alea.DiscreteEventSystems.TriangularDelay(25, 55, 40))

SM = project.NewTransition("StartActivityM", ProjectState.CanActivityMStart, ProjectState.StartActivityM, Alea.DiscreteEventSystems.DiracDelay(0))
CM = project.NewTransition("CompleteActivityM", ProjectState.CanActivityMBeCompleted, ProjectState.CompleteActivityM, Alea.DiscreteEventSystems.TriangularDelay(59, 93, 70))

SN = project.NewTransition("StartActivityN", ProjectState.CanActivityNStart, ProjectState.StartActivityN, Alea.DiscreteEventSystems.DiracDelay(0))
CN = project.NewTransition("CompleteActivityN", ProjectState.CanActivityNBeCompleted, ProjectState.CompleteActivityN, Alea.DiscreteEventSystems.TriangularDelay(25, 45, 35))

T = project.NewTransition("CompleteProject", ProjectState.CanProjectBeCompleted, ProjectState.CompleteProject, Alea.DiscreteEventSystems.DiracDelay(0))

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(project)
I1 = simulator.NewTransitionFirstFiringDate("I1", T, True)

numberOfTrials = 10000
missionTime = 1000
simulator.MonteCarloSimulation(missionTime, numberOfTrials)

print("Min\t{0:.2f}".format(I1.GetMinimumValue()))
print("Max\t{0:.2f}".format(I1.GetMaximumValue()))
print("Mean\t{0:.2f}".format(I1.GetMean()))
print("Stddev\t{0:.2f}".format(I1.GetStandardDeviation()))

plt.hist(I1.GetRecordedValues(), bins=np.arange(I1.GetMinimumValue(), I1.GetMaximumValue()+1), density=True)
plt.xlabel('Duration')
plt.ylabel('Proportion')
plt.title('Project duration distribution ({0:d} simulations)'.format(numberOfTrials))
plt.show()
