# TPK4186 - Advanced Tools for Performance Engineering
# Trolley System.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. TrolleySystemState
# 3. TrolleyGroup
# 4. Main

# 1. Required modules
# -------------------

import sys

sys.path.insert(0, '..')
import Alea.DiscreteEventSystems
import Alea.DiscreteEventSimulators

# 2. TrolleySystemState
# ---------------------

class TrolleySystemState(Alea.DiscreteEventSystems.State):
  LOAD_STATION = 0
  SECTION_1 = 1
  SECTION_2 = 2
  SECTION_3 = 3
  SECTION_4 = 4
  SECTION_5 = 5
  SECTION_6 = 7
  OFFLOAD_STATION = 8

  def __init__(self, trolleySystem):
    Alea.DiscreteEventSystems.State.__init__(self, trolleySystem)
    self.trolleyALocation = TrolleySystemState.LOAD_STATION
    self.trolleyBLocation = TrolleySystemState.LOAD_STATION
    self.trolleyAFirst = True

  @staticmethod
  def New(trolleySystem):
    return TrolleySystemState(trolleySystem)

  def Reset(self):
    self.trolleyALocation = TrolleySystemState.LOAD_STATION
    self.trolley2State = TrolleySystemState.LOAD_STATION
    self.trolleyAFirst = True


  def CanTrolleyAEnterSection1(self):
    return self.trolleyALocation==TrolleySystemState.LOAD_STATION

  def MoveTrolleyAToSection1(self):
    self.trolleyALocation = TrolleySystemState.SECTION_1

  def CanTrolleyAEnterSection2(self):
    return self.trolleyALocation==TrolleySystemState.SECTION_1

  def MoveTrolleyAToSection2(self):
    self.trolleyALocation = TrolleySystemState.SECTION_2

  def CanTrolleyAEnterSection3(self):
    return self.trolleyALocation==TrolleySystemState.SECTION_2

  def MoveTrolleyAToSection3(self):
    self.trolleyALocation = TrolleySystemState.SECTION_3

  def CanTrolleyAEnterOffloadStation(self):
    if not self.trolleyALocation==TrolleySystemState.SECTION_3 or self.trolleyBLocation==TrolleySystemState.OFFLOAD_STATION:
      return False
    if self.trolleyBLocation==TrolleySystemState.SECTION_3:
      return self.trolleyAFirst
    return True 

  def MoveTrolleyAToOffloadStation(self):
    if self.trolleyBLocation==TrolleySystemState.SECTION_3:
      self.trolleyAFirst = False
    self.trolleyALocation = TrolleySystemState.OFFLOAD_STATION

  def CanTrolleyAEnterSection4(self):
    return self.trolleyALocation==TrolleySystemState.OFFLOAD_STATION

  def MoveTrolleyAToSection4(self):
    self.trolleyALocation = TrolleySystemState.SECTION_4

  def CanTrolleyAEnterSection5(self):
    return self.trolleyALocation==TrolleySystemState.SECTION_4

  def MoveTrolleyAToSection5(self):
    self.trolleyALocation = TrolleySystemState.SECTION_5

  def CanTrolleyAEnterSection6(self):
    return self.trolleyALocation==TrolleySystemState.SECTION_5

  def MoveTrolleyAToSection6(self):
    self.trolleyALocation = TrolleySystemState.SECTION_6

  def CanTrolleyAEnterLoadStation(self):
    return self.trolleyALocation==TrolleySystemState.SECTION_6

  def MoveTrolleyAToLoadStation(self):
    self.trolleyALocation = TrolleySystemState.LOAD_STATION


  def CanTrolleyBEnterSection1(self):
    return self.trolleyBLocation==TrolleySystemState.LOAD_STATION

  def MoveTrolleyBToSection1(self):
    self.trolleyBLocation = TrolleySystemState.SECTION_1

  def CanTrolleyBEnterSection2(self):
    return self.trolleyBLocation==TrolleySystemState.SECTION_1

  def MoveTrolleyBToSection2(self):
    self.trolleyBLocation = TrolleySystemState.SECTION_2

  def CanTrolleyBEnterSection3(self):
    return self.trolleyBLocation==TrolleySystemState.SECTION_2

  def MoveTrolleyBToSection3(self):
    self.trolleyBLocation = TrolleySystemState.SECTION_3

  def CanTrolleyBEnterOffloadStation(self):
    if not self.trolleyBLocation==TrolleySystemState.SECTION_3 or self.trolleyALocation==TrolleySystemState.OFFLOAD_STATION:
      return False
    if self.trolleyALocation==TrolleySystemState.SECTION_3:
      return not self.trolleyAFirst
    return True 

  def MoveTrolleyBToOffloadStation(self):
    if self.trolleyALocation==TrolleySystemState.SECTION_3:
      self.trolleyAFirst = True
    self.trolleyBLocation = TrolleySystemState.OFFLOAD_STATION

  def CanTrolleyBEnterSection4(self):
    return self.trolleyBLocation==TrolleySystemState.OFFLOAD_STATION

  def MoveTrolleyBToSection4(self):
    self.trolleyBLocation = TrolleySystemState.SECTION_4

  def CanTrolleyBEnterSection5(self):
    return self.trolleyBLocation==TrolleySystemState.SECTION_4

  def MoveTrolleyBToSection5(self):
    self.trolleyBLocation = TrolleySystemState.SECTION_5

  def CanTrolleyBEnterSection6(self):
    return self.trolleyBLocation==TrolleySystemState.SECTION_5

  def MoveTrolleyBToSection6(self):
    self.trolleyBLocation = TrolleySystemState.SECTION_6

  def CanTrolleyBEnterLoadStation(self):
    return self.trolleyBLocation==TrolleySystemState.SECTION_6

  def MoveTrolleyBToLoadStation(self):
    self.trolleyBLocation = TrolleySystemState.LOAD_STATION


  def IsSituationDangerous(self):
    return self.trolleyALocation==TrolleySystemState.OFFLOAD_STATION and self.trolleyBLocation==TrolleySystemState.OFFLOAD_STATION

# 3. TrolleyGroup
# ----------------

minimumLoadingTime = 3
maximumLoadingTime = 5
travelTimeSection = 2
minimumOffloadingTime = 2
maximumOffloadingTime = 3

trolleySystem = Alea.DiscreteEventSystems.DiscreteEventSystem(TrolleySystemState.New)

trolleySystem.NewTransition("MoveTrolleyAToSection1", TrolleySystemState.CanTrolleyAEnterSection1, TrolleySystemState.MoveTrolleyAToSection1, Alea.DiscreteEventSystems.UniformDelay(minimumLoadingTime, maximumLoadingTime))
trolleySystem.NewTransition("MoveTrolleyAToSection2", TrolleySystemState.CanTrolleyAEnterSection2, TrolleySystemState.MoveTrolleyAToSection2, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
trolleySystem.NewTransition("MoveTrolleyAToSection3", TrolleySystemState.CanTrolleyAEnterSection3, TrolleySystemState.MoveTrolleyAToSection3, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
T = trolleySystem.NewTransition("MoveTrolleyAToOffloadStation", TrolleySystemState.CanTrolleyAEnterOffloadStation, TrolleySystemState.MoveTrolleyAToOffloadStation, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
T.SetPolicy(Alea.DiscreteEventSystems.Transition.MEMORY)
trolleySystem.NewTransition("MoveTrolleyAToSection4", TrolleySystemState.CanTrolleyAEnterSection4, TrolleySystemState.MoveTrolleyAToSection4, Alea.DiscreteEventSystems.UniformDelay(minimumOffloadingTime, maximumOffloadingTime))
trolleySystem.NewTransition("MoveTrolleyAToSection5", TrolleySystemState.CanTrolleyAEnterSection5, TrolleySystemState.MoveTrolleyAToSection5, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
trolleySystem.NewTransition("MoveTrolleyAToSection6", TrolleySystemState.CanTrolleyAEnterSection6, TrolleySystemState.MoveTrolleyAToSection6, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
trolleySystem.NewTransition("MoveTrolleyAToLoadStation", TrolleySystemState.CanTrolleyAEnterLoadStation, TrolleySystemState.MoveTrolleyAToLoadStation, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))

trolleySystem.NewTransition("MoveTrolleyBToSection1", TrolleySystemState.CanTrolleyBEnterSection1, TrolleySystemState.MoveTrolleyBToSection1, Alea.DiscreteEventSystems.UniformDelay(minimumLoadingTime, maximumLoadingTime))
trolleySystem.NewTransition("MoveTrolleyBToSection2", TrolleySystemState.CanTrolleyBEnterSection2, TrolleySystemState.MoveTrolleyBToSection2, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
trolleySystem.NewTransition("MoveTrolleyBToSection3", TrolleySystemState.CanTrolleyBEnterSection3, TrolleySystemState.MoveTrolleyBToSection3, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
T = trolleySystem.NewTransition("MoveTrolleyBToOffloadStation", TrolleySystemState.CanTrolleyBEnterOffloadStation, TrolleySystemState.MoveTrolleyBToOffloadStation, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
T.SetPolicy(Alea.DiscreteEventSystems.Transition.MEMORY)
trolleySystem.NewTransition("MoveTrolleyBToSection4", TrolleySystemState.CanTrolleyBEnterSection4, TrolleySystemState.MoveTrolleyBToSection4, Alea.DiscreteEventSystems.UniformDelay(minimumOffloadingTime, maximumOffloadingTime))
trolleySystem.NewTransition("MoveTrolleyBToSection5", TrolleySystemState.CanTrolleyBEnterSection5, TrolleySystemState.MoveTrolleyBToSection5, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
trolleySystem.NewTransition("MoveTrolleyBToSection6", TrolleySystemState.CanTrolleyBEnterSection6, TrolleySystemState.MoveTrolleyBToSection6, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))
trolleySystem.NewTransition("MoveTrolleyBToLoadStation", TrolleySystemState.CanTrolleyBEnterLoadStation, TrolleySystemState.MoveTrolleyBToLoadStation, Alea.DiscreteEventSystems.DiracDelay(travelTimeSection))

# 4. Main
# -------

simulator = Alea.DiscreteEventSimulators.DiscreteEventSimulator()
simulator.SetSystem(trolleySystem)
T = trolleySystem.LookForTransition("MoveTrolleyAToSection4")
I1 = simulator.NewTransitionFiringCount("I1", T, False)
T = trolleySystem.LookForTransition("MoveTrolleyBToSection4")
I2 = simulator.NewTransitionFiringCount("I2", T, False)
I3 = simulator.NewPredicateSatisfactionCount("I3", TrolleySystemState.IsSituationDangerous, False)

missionTime = 100
numberOfTrials = 1
simulator.MonteCarloSimulation(missionTime, numberOfTrials)

print("I1\t{0:.3e}".format(I1.GetMean()))
print("I2\t{0:.3e}".format(I2.GetMean()))
print("I3\t{0:.3e}".format(I3.GetMean()))

